@@ -13,6 +13,7 @@
 // OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
 use core::fmt;
+use core::ops::ControlFlow;
 
 /// An error that occurs during certificate validation or name validation.
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
@@ -41,6 +42,10 @@ pub enum Error {
     /// The certificate, or one of its issuers, has been revoked.
     CertRevoked,
 
+    /// The CRL is expired; i.e. the verification time is not before the time
+    /// in the CRL nextUpdate field.
+    CrlExpired,
+
     /// An end-entity certificate is being used as a CA certificate.
     EndEntityUsedAsCa,
 
@@ -93,6 +98,15 @@ pub enum Error {
     /// invalid labels.
     MalformedNameConstraint,
 
+    /// The maximum number of name constraint comparisons has been reached.
+    MaximumNameConstraintComparisonsExceeded,
+
+    /// The maximum number of internal path building calls has been reached. Path complexity is too great.
+    MaximumPathBuildCallsExceeded,
+
+    /// The path search was terminated because it became too deep.
+    MaximumPathDepthExceeded,
+
     /// The maximum number of signature checks has been reached. Path complexity is too great.
     MaximumSignatureChecksExceeded,
 
@@ -110,9 +124,15 @@ pub enum Error {
     /// does not match the algorithm in the signature of the certificate.
     SignatureAlgorithmMismatch,
 
+    /// Trailing data was found while parsing DER-encoded input for the named type.
+    TrailingData(DerTypeId),
+
     /// A valid issuer for the certificate could not be found.
     UnknownIssuer,
 
+    /// The certificate's revocation status could not be determined.
+    UnknownRevocationStatus,
+
     /// The certificate is not a v3 X.509 certificate.
     ///
     /// This error may be also reported if the certificate version field
@@ -122,6 +142,10 @@ pub enum Error {
     /// The certificate contains an unsupported critical extension.
     UnsupportedCriticalExtension,
 
+    /// The CRL contains an issuing distribution point with no distribution point name,
+    /// or a distribution point name relative to an issuer.
+    UnsupportedCrlIssuingDistributionPoint,
+
     /// The CRL is not a v2 X.509 CRL.
     ///
     /// The RFC 5280 web PKI profile mandates only version 2 be used. See section
@@ -136,9 +160,15 @@ pub enum Error {
     /// The CRL contains unsupported "indirect" entries.
     UnsupportedIndirectCrl,
 
+    /// The `ServerName` contained an unsupported type of value.
+    UnsupportedNameType,
+
     /// The revocation reason is not in the set of supported revocation reasons.
     UnsupportedRevocationReason,
 
+    /// The CRL is partitioned by revocation reasons.
+    UnsupportedRevocationReasonsPartitioning,
+
     /// The signature algorithm for a signature over a CRL is not in the set of supported
     /// signature algorithms given.
     UnsupportedCrlSignatureAlgorithm,
@@ -185,51 +215,80 @@ impl Error {
     pub(crate) fn rank(&self) -> u32 {
         match &self {
             // Errors related to certificate validity
-            Error::CertNotValidYet | Error::CertExpired => 27,
-            Error::CertNotValidForName => 26,
-            Error::CertRevoked => 25,
-            Error::InvalidCrlSignatureForPublicKey | Error::InvalidSignatureForPublicKey => 24,
-            Error::SignatureAlgorithmMismatch => 23,
-            Error::RequiredEkuNotFound => 22,
-            Error::NameConstraintViolation => 21,
-            Error::PathLenConstraintViolated => 20,
-            Error::CaUsedAsEndEntity | Error::EndEntityUsedAsCa => 19,
-            Error::IssuerNotCrlSigner => 18,
+            Error::CertNotValidYet | Error::CertExpired => 290,
+            Error::CertNotValidForName => 280,
+            Error::CertRevoked | Error::UnknownRevocationStatus | Error::CrlExpired => 270,
+            Error::InvalidCrlSignatureForPublicKey | Error::InvalidSignatureForPublicKey => 260,
+            Error::SignatureAlgorithmMismatch => 250,
+            Error::RequiredEkuNotFound => 240,
+            Error::NameConstraintViolation => 230,
+            Error::PathLenConstraintViolated => 220,
+            Error::CaUsedAsEndEntity | Error::EndEntityUsedAsCa => 210,
+            Error::IssuerNotCrlSigner => 200,
 
             // Errors related to supported features used in an invalid way.
-            Error::InvalidCertValidity => 17,
-            Error::InvalidNetworkMaskConstraint => 16,
-            Error::InvalidSerialNumber => 15,
-            Error::InvalidCrlNumber => 14,
+            Error::InvalidCertValidity => 190,
+            Error::InvalidNetworkMaskConstraint => 180,
+            Error::InvalidSerialNumber => 170,
+            Error::InvalidCrlNumber => 160,
 
             // Errors related to unsupported features.
             Error::UnsupportedCrlSignatureAlgorithmForPublicKey
-            | Error::UnsupportedSignatureAlgorithmForPublicKey => 13,
-            Error::UnsupportedCrlSignatureAlgorithm | Error::UnsupportedSignatureAlgorithm => 12,
-            Error::UnsupportedCriticalExtension => 11,
-            Error::UnsupportedCertVersion => 11,
-            Error::UnsupportedCrlVersion => 10,
-            Error::UnsupportedDeltaCrl => 9,
-            Error::UnsupportedIndirectCrl => 8,
-            Error::UnsupportedRevocationReason => 7,
+            | Error::UnsupportedSignatureAlgorithmForPublicKey => 150,
+            Error::UnsupportedCrlSignatureAlgorithm | Error::UnsupportedSignatureAlgorithm => 140,
+            Error::UnsupportedCriticalExtension => 130,
+            Error::UnsupportedCertVersion => 130,
+            Error::UnsupportedCrlVersion => 120,
+            Error::UnsupportedDeltaCrl => 110,
+            Error::UnsupportedIndirectCrl => 100,
+            Error::UnsupportedNameType => 95,
+            Error::UnsupportedRevocationReason => 90,
+            Error::UnsupportedRevocationReasonsPartitioning => 80,
+            Error::UnsupportedCrlIssuingDistributionPoint => 70,
+            Error::MaximumPathDepthExceeded => 61,
 
             // Errors related to malformed data.
-            Error::MalformedDnsIdentifier => 6,
-            Error::MalformedNameConstraint => 5,
-            Error::MalformedExtensions => 4,
-            Error::ExtensionValueInvalid => 3,
+            Error::MalformedDnsIdentifier => 60,
+            Error::MalformedNameConstraint => 50,
+            Error::MalformedExtensions | Error::TrailingData(_) => 40,
+            Error::ExtensionValueInvalid => 30,
 
             // Generic DER errors.
-            Error::BadDerTime => 2,
-            Error::BadDer => 1,
+            Error::BadDerTime => 20,
+            Error::BadDer => 10,
 
-            // Special case error - not subject to ranking.
+            // Special case errors - not subject to ranking.
             Error::MaximumSignatureChecksExceeded => 0,
+            Error::MaximumPathBuildCallsExceeded => 0,
+            Error::MaximumNameConstraintComparisonsExceeded => 0,
 
             // Default catch all error - should be renamed in the future.
             Error::UnknownIssuer => 0,
         }
     }
+
+    /// Returns true for errors that should be considered fatal during path building. Errors of
+    /// this class should halt any further path building and be returned immediately.
+    #[inline]
+    pub(crate) fn is_fatal(&self) -> bool {
+        matches!(
+            self,
+            Error::MaximumSignatureChecksExceeded
+                | Error::MaximumPathBuildCallsExceeded
+                | Error::MaximumNameConstraintComparisonsExceeded
+        )
+    }
+}
+
+impl From<Error> for ControlFlow<Error, Error> {
+    fn from(value: Error) -> Self {
+        match value {
+            // If an error is fatal, we've exhausted the potential for continued search.
+            err if err.is_fatal() => Self::Break(err),
+            // Otherwise we've rejected one candidate chain, but may continue to search for others.
+            err => Self::Continue(err),
+        }
+    }
 }
 
 impl fmt::Display for Error {
@@ -239,11 +298,37 @@ impl fmt::Display for Error {
 }
 
 #[cfg(feature = "std")]
-#[cfg_attr(docsrs, doc(cfg(feature = "std")))]
 impl ::std::error::Error for Error {}
 
-impl From<untrusted::EndOfInput> for Error {
-    fn from(_: untrusted::EndOfInput) -> Self {
-        Error::BadDer
-    }
+/// Trailing data was found while parsing DER-encoded input for the named type.
+#[allow(missing_docs)]
+#[non_exhaustive]
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub enum DerTypeId {
+    BitString,
+    Bool,
+    Certificate,
+    CertificateExtensions,
+    CertificateTbsCertificate,
+    CertRevocationList,
+    CertRevocationListExtension,
+    CrlDistributionPoint,
+    CommonNameInner,
+    CommonNameOuter,
+    DistributionPointName,
+    Extension,
+    GeneralName,
+    RevocationReason,
+    Signature,
+    SignatureAlgorithm,
+    SignedData,
+    SubjectPublicKeyInfo,
+    Time,
+    TrustAnchorV1,
+    TrustAnchorV1TbsCertificate,
+    U8,
+    RevokedCertificate,
+    RevokedCertificateExtension,
+    RevokedCertEntry,
+    IssuingDistributionPoint,
 }
